using System;
using System.Collections.Generic;
using System.Text;

namespace PrimeNumbers
{
    partial class PNTable
    {
        protected uint[] arr = null;
        protected int size = 0;
        protected Int64 max_number = 0;

        public Int64 get_index_by_number(Int64 num)
        {
            Int64 a = num - first_number;
            Int64 d = a / period_sum;
            Int64 m = a % period_sum;

            int sum = 0;
            int i;

            for ( i = 0; i < period.Length; i++ )
            {
                if ( m < sum ) break;
                sum += period[i];
            }
            
            i--;

            return d*period.Length + i;
        }

        public Int64 get_number_by_index(Int64 idx)
        {
            Int64 a = idx / period.Length;
            Int64 b = idx % period.Length;

            Int64 sum = 0;
            for ( int i = 0; i < b; i++ )
                sum += period[i];

            return a * period_sum + sum + first_number;
        }

        protected void reset_bit(Int64 num)
        {
            Int64 i = num / 32;
            Int64 p = num % 32;

            arr[i] = arr[i] & (uint)(~(1 << (int)p));
        }

        protected bool get_bit(Int64 num)
        {
            Int64 i = num / 32;
            Int64 p = num % 32;

            if ( (arr[i] & (uint)(1 << (int)p)) != 0 )
                return true;
            return false;
        }

        public PNTable() : this(1024 * 1024 * 1024)
        {
        }

        public PNTable(int sz)
        {
            size = sz;
            arr = new uint[size / 4];

            Int64 i;
            
            for ( i = 0; i < arr.Length; i++ )
                arr[i] = 0xffffffff;

            max_number = get_number_by_index((Int64)arr.Length * 32);
            Int64 bit_numbers = (Int64)arr.Length * (Int64)32;

            Int64[] inc_tab = new Int64[period.Length+1];

            for ( i = 0; i < (Int64)arr.Length*(Int64)32; i++ )
            {
                if ( get_bit(i) )
                {
                    Int64 num = get_number_by_index(i);

                    if ( num*num > max_number )
                        break;

                    Int64 j;
                    Int64 idx = 0;

                    inc_tab[idx++] = i;

                    for ( j = num * first_number; j < max_number; j += num )
                    {
                        if ( !is_base(j) )
                        {
                            inc_tab[idx++] = get_index_by_number(j);
                            if ( idx == inc_tab.Length )
                                break;
                        }
                    }

                    for ( j = 1; j < inc_tab.Length; j++ )
                        inc_tab[j-1] = inc_tab[j] - inc_tab[j-1];

                    idx = 0;
                    j = i + inc_tab[idx++];
                    while ( j < bit_numbers )
                    {
                        reset_bit(j);
                        j += inc_tab[idx++];
                        idx %= period.Length;
                    }
                }
            } 
        }

        protected bool is_base(Int64 num)
        {
            for ( int i = 0; i < base_primes.Length; i++ )
                if ( num % base_primes[i] == 0 )
                    return true;
            return false;
        }

        public Int64 MaxNumber
        {
            get { return max_number; }
        }

        public IEnumerable<Int64> Table
        {
            get 
            {
                Int64 i;
                for ( i = 0; i < base_primes.Length; i++ )
                    yield return base_primes[i];

                Int64 bit_numbers = (Int64)arr.Length * (Int64)32;
                Int64 ps = 0;
                Int64 cur_num = first_number;
                Int64 cur_bit_pos = 0;

                while ( cur_bit_pos < bit_numbers )
                {
                    if ( get_bit(cur_bit_pos) )
                        yield return cur_num;
                    cur_bit_pos++;
                    cur_num += period[ps++];
                    ps %= period.Length;
                }
            }
        }
    }
}
